# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import dataclasses
import datetime
import random
from typing import TYPE_CHECKING

from twisted.trial import unittest

from buildbot.data import base
from buildbot.data import resultspec
from buildbot.data.resultspec import NoneComparator
from buildbot.data.resultspec import ReverseComparator

if TYPE_CHECKING:
    from typing import ClassVar
    from typing import Sequence


class ResultSpecMKListMixin:
    @staticmethod
    def mkdata(fld: Sequence[str] | str, *values):
        if isinstance(fld, str):
            return [{fld: val} for val in values]

        return [dict(zip(fld, val)) for val in values]


class ResultSpecMKDataclassMixin(ResultSpecMKListMixin):
    dataclasses_cache: ClassVar[dict[str, type]] = {}

    @classmethod
    def _get_dataclass(cls, fields: Sequence[str]) -> type:
        """
        Re-use runtime dataclasses so comparison work
        """
        class_key = f"ResultSpecMKDataclassMixin_{'_'.join(fields)}"
        if class_key not in cls.dataclasses_cache:
            test_cls = dataclasses.make_dataclass(class_key, fields)
            cls.dataclasses_cache[class_key] = test_cls

        return cls.dataclasses_cache[class_key]

    @staticmethod
    def mkdata(fld: Sequence[str] | str, *values):
        if isinstance(fld, str):
            fields = [fld]
        else:
            fields = sorted(fld)

        test_cls = ResultSpecMKDataclassMixin._get_dataclass(fields)

        return [test_cls(**item) for item in ResultSpecMKListMixin.mkdata(fld, *values)]


class FilterTestMixin:
    @staticmethod
    def mkdata(fld, *values) -> list:
        raise NotImplementedError()

    def test_eq(self):
        f = resultspec.Filter('num', 'eq', [10])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10))), self.mkdata('num', 10))

    def test_eq_plural(self):
        f = resultspec.Filter('num', 'eq', [10, 15, 20])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10, 15))), self.mkdata('num', 10, 15))

    def test_ne(self):
        f = resultspec.Filter('num', 'ne', [10])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10))), self.mkdata('num', 5))

    def test_ne_plural(self):
        f = resultspec.Filter('num', 'ne', [10, 15, 20])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10, 15))), self.mkdata('num', 5))

    def test_lt(self):
        f = resultspec.Filter('num', 'lt', [10])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10, 15))), self.mkdata('num', 5))

    def test_le(self):
        f = resultspec.Filter('num', 'le', [10])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10, 15))), self.mkdata('num', 5, 10))

    def test_gt(self):
        f = resultspec.Filter('num', 'gt', [10])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10, 15))), self.mkdata('num', 15))

    def test_ge(self):
        f = resultspec.Filter('num', 'ge', [10])
        self.assertEqual(list(f.apply(self.mkdata('num', 5, 10, 15))), self.mkdata('num', 10, 15))

    def test_contains(self):
        f = resultspec.Filter('num', 'contains', [10])
        self.assertEqual(
            list(f.apply(self.mkdata('num', [5, 1], [10, 1], [15, 1]))), self.mkdata('num', [10, 1])
        )

    def test_contains_plural(self):
        f = resultspec.Filter('num', 'contains', [10, 5])
        self.assertEqual(
            list(f.apply(self.mkdata('num', [5, 1], [10, 1], [15, 1]))),
            self.mkdata('num', [5, 1], [10, 1]),
        )


class FilterList(unittest.TestCase, ResultSpecMKListMixin, FilterTestMixin):
    pass


class FilterDataclass(unittest.TestCase, ResultSpecMKDataclassMixin, FilterTestMixin):
    pass


class ResultSpecTestMixin:
    @staticmethod
    def mkdata(fld, *values) -> list:
        raise NotImplementedError()

    def assertListResultEqual(self, a, b):
        self.assertIsInstance(a, base.ListResult)
        self.assertIsInstance(b, base.ListResult)
        self.assertEqual(a, b)

    def test_apply_None(self):
        self.assertEqual(resultspec.ResultSpec().apply(None), None)

    def test_apply_details_fields(self):
        data = {"name": 'clyde', "id": 14, "favcolor": 'red'}
        self.assertEqual(resultspec.ResultSpec(fields=['name']).apply(data), {"name": 'clyde'})
        self.assertEqual(
            resultspec.ResultSpec(fields=['name', 'id']).apply(data), {"name": 'clyde', "id": 14}
        )

    def test_apply_collection_fields(self):
        data = self.mkdata(('a', 'b', 'c'), (1, 11, 111), (2, 22, 222))
        self.assertEqual(resultspec.ResultSpec(fields=['a']).apply(data), [{'a': 1}, {'a': 2}])
        self.assertEqual(
            resultspec.ResultSpec(fields=['a', 'c']).apply(data),
            [{'a': a, 'c': c} for a, c in [(1, 111), (2, 222)]],
        )

    def test_apply_ordering(self):
        data = self.mkdata('name', 'albert', 'bruce', 'cedric', 'dwayne')
        exp = self.mkdata('name', 'albert', 'bruce', 'cedric', 'dwayne')
        random.shuffle(data)
        self.assertEqual(resultspec.ResultSpec(order=['name']).apply(data), exp)
        self.assertEqual(resultspec.ResultSpec(order=['-name']).apply(data), list(reversed(exp)))

    def test_apply_ordering_multi(self):
        data = self.mkdata(
            ('fn', 'ln'),
            ('cedric', 'willis'),
            ('albert', 'engelbert'),
            ('bruce', 'willis'),
            ('dwayne', 'montague'),
        )
        exp = base.ListResult(
            self.mkdata(
                ('fn', 'ln'),
                ('albert', 'engelbert'),
                ('dwayne', 'montague'),
                ('bruce', 'willis'),
                ('cedric', 'willis'),
            ),
            total=4,
        )
        random.shuffle(data)
        self.assertListResultEqual(resultspec.ResultSpec(order=['ln', 'fn']).apply(data), exp)
        exp = base.ListResult(
            self.mkdata(
                ('fn', 'ln'),
                ('cedric', 'willis'),
                ('bruce', 'willis'),
                ('dwayne', 'montague'),
                ('albert', 'engelbert'),
            ),
            total=4,
        )
        self.assertListResultEqual(resultspec.ResultSpec(order=['-ln', '-fn']).apply(data), exp)

    def test_apply_filter(self):
        data = self.mkdata('name', 'albert', 'bruce', 'cedric', 'dwayne')
        f = resultspec.Filter(field='name', op='gt', values=['bruce'])
        self.assertListResultEqual(
            resultspec.ResultSpec(filters=[f]).apply(data),
            base.ListResult(self.mkdata('name', 'cedric', 'dwayne'), total=2),
        )
        f2 = resultspec.Filter(field='name', op='le', values=['cedric'])
        self.assertListResultEqual(
            resultspec.ResultSpec(filters=[f, f2]).apply(data),
            base.ListResult(self.mkdata('name', 'cedric'), total=1),
        )

    def test_apply_missing_fields(self):
        data = self.mkdata(
            ('fn', 'ln'),
            ('cedric', 'willis'),
            ('albert', 'engelbert'),
            ('bruce', 'willis'),
            ('dwayne', 'montague'),
        )
        resultspec.ResultSpec(fields=['fn'], order=['ln']).apply(data)

    def test_sort_null_datetimefields(self):
        data = self.mkdata(('fn', 'ln'), ('albert', datetime.datetime(1, 1, 1)), ('cedric', None))

        exp = self.mkdata(('fn', 'ln'), ('cedric', None), ('albert', datetime.datetime(1, 1, 1)))

        self.assertListResultEqual(
            resultspec.ResultSpec(order=['ln']).apply(data), base.ListResult(exp, total=2)
        )

    def do_test_pagination(self, bareList):
        data = self.mkdata('x', *list(range(101, 131)))
        if not bareList:
            data = base.ListResult(data)
            data.offset = None
            data.total = len(data)
            data.limit = None
        self.assertListResultEqual(
            resultspec.ResultSpec(offset=0).apply(data),
            base.ListResult(self.mkdata('x', *list(range(101, 131))), offset=0, total=30),
        )
        self.assertListResultEqual(
            resultspec.ResultSpec(offset=10).apply(data),
            base.ListResult(self.mkdata('x', *list(range(111, 131))), offset=10, total=30),
        )
        self.assertListResultEqual(
            resultspec.ResultSpec(offset=10, limit=10).apply(data),
            base.ListResult(
                self.mkdata('x', *list(range(111, 121))), offset=10, total=30, limit=10
            ),
        )
        self.assertListResultEqual(
            resultspec.ResultSpec(offset=20, limit=15).apply(data),
            base.ListResult(
                self.mkdata('x', *list(range(121, 131))), offset=20, total=30, limit=15
            ),
        )  # off the end

    def test_pagination_bare_list(self):
        return self.do_test_pagination(bareList=True)

    def test_pagination_ListResult(self):
        return self.do_test_pagination(bareList=False)

    def test_pagination_prepaginated(self):
        data = base.ListResult(self.mkdata('x', *list(range(10, 20))))
        data.offset = 10
        data.total = 30
        data.limit = 10
        self.assertListResultEqual(
            # ResultSpec has its offset/limit fields cleared
            resultspec.ResultSpec().apply(data),
            base.ListResult(self.mkdata('x', *list(range(10, 20))), offset=10, total=30, limit=10),
        )

    def test_pagination_prepaginated_without_clearing_resultspec(self):
        data = base.ListResult(self.mkdata('x', *list(range(10, 20))))
        data.offset = 10
        data.limit = 10
        # ResultSpec does not have its offset/limit fields cleared - this is
        # detected as an assertion failure
        with self.assertRaises(AssertionError):
            resultspec.ResultSpec(offset=10, limit=20).apply(data)

    def test_endpoint_returns_total_without_applying_filters(self):
        data = base.ListResult(self.mkdata('x', *list(range(10, 20))))
        data.total = 99
        # apply doesn't want to get a total with filters still outstanding
        f = resultspec.Filter(field='x', op='gt', values=[23])
        with self.assertRaises(AssertionError):
            resultspec.ResultSpec(filters=[f]).apply(data)

    def test_popProperties(self):
        expected = ['prop1', 'prop2']
        rs = resultspec.ResultSpec(properties=[resultspec.Property(b'property', 'eq', expected)])
        self.assertEqual(len(rs.properties), 1)
        self.assertEqual(rs.popProperties(), expected)
        self.assertEqual(len(rs.properties), 0)

    def test_popFilter(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', [10]),
                resultspec.Filter('foo', 'gt', [5]),
                resultspec.Filter('base', 'ne', [20]),
            ]
        )
        self.assertEqual(rs.popFilter('baz', 'lt'), None)  # no match
        self.assertEqual(rs.popFilter('foo', 'eq'), [10])
        self.assertEqual(len(rs.filters), 2)

    def test_popBooleanFilter(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', [True]),
                resultspec.Filter('bar', 'ne', [False]),
            ]
        )
        self.assertEqual(rs.popBooleanFilter('foo'), True)
        self.assertEqual(rs.popBooleanFilter('bar'), True)
        self.assertEqual(len(rs.filters), 0)

    def test_popStringFilter(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', ['foo']),
            ]
        )
        self.assertEqual(rs.popStringFilter('foo'), 'foo')

    def test_popStringFilterSeveral(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', ['foo', 'bar']),
            ]
        )
        self.assertEqual(rs.popStringFilter('foo'), None)

    def test_popIntegerFilter(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', ['12']),
            ]
        )
        self.assertEqual(rs.popIntegerFilter('foo'), 12)

    def test_popIntegerFilterSeveral(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', ['12', '13']),
            ]
        )
        self.assertEqual(rs.popIntegerFilter('foo'), None)

    def test_popIntegerFilterNotInt(self):
        rs = resultspec.ResultSpec(
            filters=[
                resultspec.Filter('foo', 'eq', ['bar']),
            ]
        )
        with self.assertRaises(ValueError):
            rs.popIntegerFilter('foo')

    def test_removeOrder(self):
        rs = resultspec.ResultSpec(order=['foo', '-bar'])
        rs.removeOrder()
        self.assertEqual(rs.order, None)

    def test_popField(self):
        rs = resultspec.ResultSpec(fields=['foo', 'bar'])
        self.assertTrue(rs.popField('foo'))
        self.assertEqual(rs.fields, ['bar'])

    def test_popField_not_present(self):
        rs = resultspec.ResultSpec(fields=['foo', 'bar'])
        self.assertFalse(rs.popField('nosuch'))
        self.assertEqual(rs.fields, ['foo', 'bar'])


class ResultSpecList(unittest.TestCase, ResultSpecMKListMixin, ResultSpecTestMixin):
    def test_apply_missing_fields(self):
        # note that the REST interface catches this with a nicer error message
        with self.assertRaises(KeyError):
            super().test_apply_missing_fields()


class ResultSpecDataclass(unittest.TestCase, ResultSpecMKDataclassMixin, ResultSpecTestMixin):
    def test_apply_missing_fields(self):
        with self.assertRaises(TypeError):
            super().test_apply_missing_fields()

    def test_apply_collection_fields(self):
        with self.assertRaises(TypeError):
            super().test_apply_collection_fields()


class ComparatorTestMixin:
    @staticmethod
    def mkdata(fld, *values) -> list:
        raise NotImplementedError()

    def test_noneComparator(self):
        self.assertNotEqual(NoneComparator(None), NoneComparator(datetime.datetime(1, 1, 1)))
        self.assertNotEqual(NoneComparator(datetime.datetime(1, 1, 1)), NoneComparator(None))
        self.assertLess(NoneComparator(None), NoneComparator(datetime.datetime(1, 1, 1)))
        self.assertGreater(NoneComparator(datetime.datetime(1, 1, 1)), NoneComparator(None))
        self.assertLess(
            NoneComparator(datetime.datetime(1, 1, 1)), NoneComparator(datetime.datetime(1, 1, 2))
        )
        self.assertEqual(
            NoneComparator(datetime.datetime(1, 1, 1)), NoneComparator(datetime.datetime(1, 1, 1))
        )
        self.assertGreater(
            NoneComparator(datetime.datetime(1, 1, 2)), NoneComparator(datetime.datetime(1, 1, 1))
        )
        self.assertEqual(NoneComparator(None), NoneComparator(None))

    def test_noneComparison(self):
        noneInList = ["z", None, None, "q", "a", None, "v"]
        sortedList = sorted(noneInList, key=NoneComparator)
        self.assertEqual(sortedList, [None, None, None, "a", "q", "v", "z"])

    def test_reverseComparator(self):
        reverse35 = ReverseComparator(35)
        reverse36 = ReverseComparator(36)
        self.assertEqual(reverse35, reverse35)
        self.assertNotEqual(reverse35, reverse36)
        self.assertLess(reverse36, reverse35)
        self.assertGreater(reverse35, reverse36)
        self.assertLess(reverse36, reverse35)

    def test_reverseComparison(self):
        nums = [1, 2, 3, 4, 5]
        nums.sort(key=ReverseComparator)
        self.assertEqual(nums, [5, 4, 3, 2, 1])

    def test_reverseComparisonWithNone(self):
        noneInList = ["z", None, None, "q", "a", None, "v"]
        sortedList = sorted(noneInList, key=lambda x: ReverseComparator(NoneComparator(x)))
        self.assertEqual(sortedList, ["z", "v", "q", "a", None, None, None])


class ComparatorList(unittest.TestCase, ResultSpecMKListMixin, ComparatorTestMixin):
    pass


class ComparatorDataclass(unittest.TestCase, ResultSpecMKDataclassMixin, ComparatorTestMixin):
    pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.trial.unittest import TestCase

from buildbot_worker.test.util import command

if TYPE_CHECKING:
    from typing import Any
    from typing import TypeVar

    _T = TypeVar('_T')


# TODO: Remove unused?
class SourceCommandTestMixin(command.CommandTestMixin):
    """
    Support for testing Source Commands; an extension of CommandTestMixin
    """

    def make_command(
        self,
        cmdclass: type[command.CommandType],
        args: dict[str, Any],
        makedirs: bool = False,
        initial_sourcedata: str = '',
    ) -> command.CommandType:
        """
        Same as the parent class method, but this also adds some source-specific
        patches:

        * writeSourcedata - writes to self.sourcedata (self is the TestCase)
        * readSourcedata - reads from self.sourcedata
        * doClobber - invokes RunProcess(0, ['clobber', DIRECTORY])
        * doCopy - invokes RunProcess(0, ['copy', cmd.srcdir, cmd.workdir])
        """

        cmd = super().make_command(cmdclass, args, makedirs)

        # note that these patches are to an *instance*, not a class, so there
        # is no need to use self.patch() to reverse them

        self.sourcedata = initial_sourcedata

        def readSourcedata() -> str:
            if self.sourcedata is None:
                raise OSError("File not found")
            return self.sourcedata

        cmd.readSourcedata = readSourcedata  # type: ignore[attr-defined]

        def writeSourcedata(res: _T) -> _T:
            self.sourcedata = cmd.sourcedata  # type: ignore[attr-defined]
            return res

        cmd.writeSourcedata = writeSourcedata  # type: ignore[attr-defined]

        return cmd

    def check_sourcedata(self, _: _T, expected_sourcedata: None) -> _T:
        """
        Assert that the sourcedata (from the patched functions - see
        make_command) is correct.  Use this as a deferred callback.
        """
        assert isinstance(self, TestCase)
        self.assertEqual(self.sourcedata, expected_sourcedata)
        return _



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.trial import unittest

from buildbot import config
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.schedulers import base
from buildbot.schedulers import dependent
from buildbot.test import fakedb
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import scheduler
from buildbot.test.util.state import StateTestMixin

SUBMITTED_AT_TIME = 111111111
COMPLETE_AT_TIME = 222222222
OBJECTID = 33
SCHEDULERID = 133
UPSTREAM_NAME = 'uppy'


class Dependent(scheduler.SchedulerMixin, TestReactorMixin, StateTestMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        yield self.setUpScheduler()

    @defer.inlineCallbacks
    def makeScheduler(self, upstream=None):
        # build a fake upstream scheduler
        class Upstream(base.ReconfigurableBaseScheduler):
            def __init__(self, name):
                super().__init__(name=name, builderNames=['a'])

        if not upstream:
            upstream = Upstream(UPSTREAM_NAME)

        sched = dependent.Dependent(name='n', builderNames=['b'], upstream=upstream)
        yield self.attachScheduler(
            sched, OBJECTID, SCHEDULERID, overrideBuildsetMethods=True, createBuilderDB=True
        )

        return sched

    @defer.inlineCallbacks
    def assertBuildsetSubscriptions(self, bsids=None):
        yield self.assert_state(OBJECTID, upstream_bsids=bsids)

    # tests

    # NOTE: these tests take advantage of the fact that all of the fake
    # scheduler operations are synchronous, and thus do not return a Deferred.
    # The Deferred from trigger() is completely processed before this test
    # method returns.

    @defer.inlineCallbacks
    def test_constructor_string_arg(self):
        with self.assertRaises(config.ConfigErrors):
            yield self.makeScheduler(upstream='foo')

    @defer.inlineCallbacks
    def test_activate(self):
        sched = yield self.makeScheduler()
        yield self.master.startService()

        self.assertEqual(
            sorted([q.filter for q in sched.master.mq.qrefs]),
            [
                (
                    'buildsets',
                    None,
                    'complete',
                ),
                (
                    'buildsets',
                    None,
                    'new',
                ),
                ('schedulers', '133', 'updated'),
            ],
        )

        yield sched.deactivate()

        self.assertEqual(
            [q.filter for q in sched.master.mq.qrefs], [('schedulers', '133', 'updated')]
        )

    def sendBuildsetMessage(self, scheduler_name=None, results=-1, complete=False):
        """Call callConsumer with a buildset message.  Most of the values here
        are hard-coded to correspond to those in do_test."""
        msg = {
            "bsid": 44,
            "sourcestamps": [],  # blah blah blah
            "submitted_at": SUBMITTED_AT_TIME,
            "complete": complete,
            "complete_at": COMPLETE_AT_TIME if complete else None,
            "external_idstring": None,
            "reason": 'Because',
            "results": results if complete else -1,
            "parent_buildid": None,
            "parent_relationship": None,
        }
        if not complete:
            msg['scheduler'] = scheduler_name
        self.master.mq.callConsumer(('buildsets', '44', 'complete' if complete else 'new'), msg)

    @defer.inlineCallbacks
    def do_test(self, scheduler_name, expect_subscription, results, expect_buildset):
        """Test the dependent scheduler by faking a buildset and subsequent
        completion from an upstream scheduler.

        @param scheduler_name: upstream scheduler's name
        @param expect_subscription: whether to expect the dependent to
            subscribe to the buildset
        @param results: results of the upstream scheduler's buildset
        @param expect_buidlset: whether to expect the dependent to generate
            a new buildset in response
        """

        yield self.master.db.insert_test_data([
            fakedb.ObjectState(objectid=OBJECTID, name='upstream_bsids', value_json='[]'),
            fakedb.Object(id=OBJECTID),
        ])

        yield self.makeScheduler()
        yield self.master.startService()

        # announce a buildset with a matching name..
        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(
                id=93,
                revision='555',
                branch='master',
                project='proj',
                repository='repo',
                codebase='cb',
            ),
            fakedb.Buildset(
                id=44,
                submitted_at=SUBMITTED_AT_TIME,
                complete=False,
                complete_at=None,
                external_idstring=None,
                reason='Because',
                results=-1,
            ),
            fakedb.BuildsetSourceStamp(buildsetid=44, sourcestampid=93),
        ])
        self.sendBuildsetMessage(scheduler_name=scheduler_name, complete=False)
        yield self.master.mq._deferwaiter.wait()

        # check whether scheduler is subscribed to that buildset
        if expect_subscription:
            yield self.assertBuildsetSubscriptions([44])
        else:
            yield self.assertBuildsetSubscriptions([])

        # pretend that the buildset is finished
        yield self.master.db.buildsets.completeBuildset(bsid=44, results=results)
        self.sendBuildsetMessage(results=results, complete=True)

        # and check whether a buildset was added in response
        if expect_buildset:
            self.assertEqual(
                self.addBuildsetCalls,
                [
                    (
                        'addBuildsetForSourceStamps',
                        {
                            "builderNames": None,  # defaults
                            "external_idstring": None,
                            "properties": None,
                            "reason": 'downstream',
                            "sourcestamps": [93],
                        },
                    ),
                ],
            )
        else:
            self.assertEqual(self.addBuildsetCalls, [])

    def test_related_buildset_SUCCESS(self):
        return self.do_test(UPSTREAM_NAME, True, SUCCESS, True)

    def test_related_buildset_WARNINGS(self):
        return self.do_test(UPSTREAM_NAME, True, WARNINGS, True)

    def test_related_buildset_FAILURE(self):
        return self.do_test(UPSTREAM_NAME, True, FAILURE, False)

    def test_unrelated_buildset(self):
        return self.do_test('unrelated', False, SUCCESS, False)

    @defer.inlineCallbacks
    def test_getUpstreamBuildsets_missing(self):
        sched = yield self.makeScheduler()

        # insert some state, with more bsids than exist
        yield self.master.db.insert_test_data([
            fakedb.SourceStamp(id=1234),
            fakedb.Buildset(id=11),
            fakedb.Buildset(id=13),
            fakedb.BuildsetSourceStamp(buildsetid=13, sourcestampid=1234),
            fakedb.Object(id=OBJECTID),
            fakedb.ObjectState(objectid=OBJECTID, name='upstream_bsids', value_json='[11,12,13]'),
        ])

        # check return value (missing 12)
        self.assertEqual(
            (yield sched._getUpstreamBuildsets()), [(11, [], False, -1), (13, [1234], False, -1)]
        )

        # and check that it wrote the correct value back to the state
        yield self.assert_state(OBJECTID, upstream_bsids=[11, 13])

    @defer.inlineCallbacks
    def test_enabled_callback(self):
        sched = yield self.makeScheduler()
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)
        expectedValue = not sched.enabled
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, expectedValue)

    @defer.inlineCallbacks
    def test_disabled_activate(self):
        sched = yield self.makeScheduler()
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.activate()
        self.assertEqual(r, None)

    @defer.inlineCallbacks
    def test_disabled_deactivate(self):
        sched = yield self.makeScheduler()
        yield sched._enabledCallback(None, {'enabled': not sched.enabled})
        self.assertEqual(sched.enabled, False)
        r = yield sched.deactivate()
        self.assertEqual(r, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import datetime
import json
from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import buildsets
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import UTC
from buildbot.util import datetime2epoch
from buildbot.util import epoch2datetime


class Tests(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

        self.now = 9272359
        self.reactor.advance(self.now)

        # set up a sourcestamp with id 234 for use below
        yield self.db.insert_test_data([
            fakedb.SourceStamp(id=234),
            fakedb.Builder(id=1, name='bldr1'),
            fakedb.Builder(id=2, name='bldr2'),
        ])

    @defer.inlineCallbacks
    def test_addBuildset_getBuildset(self):
        bsid, _ = yield self.db.buildsets.addBuildset(
            sourcestamps=[234],
            reason='because',
            properties={},
            builderids=[1],
            external_idstring='extid',
            waited_for=False,
        )

        # TODO: verify buildrequests too
        bsdict = yield self.db.buildsets.getBuildset(bsid)
        self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            bsdict,
            buildsets.BuildSetModel(
                bsid=bsid,
                external_idstring='extid',
                reason='because',
                submitted_at=datetime.datetime(1970, 4, 18, 7, 39, 19, tzinfo=UTC),
                results=-1,
                sourcestamps=[234],
            ),
        )

    @defer.inlineCallbacks
    def test_addBuildset_getBuildset_explicit_submitted_at(self):
        bsid_brids = yield self.db.buildsets.addBuildset(
            sourcestamps=[234],
            reason='because',
            properties={},
            builderids=[1],
            external_idstring='extid',
            submitted_at=epoch2datetime(8888888),
            waited_for=False,
        )
        bsdict = yield self.db.buildsets.getBuildset(bsid_brids[0])

        self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            bsdict,
            buildsets.BuildSetModel(
                bsid=bsid_brids[0],
                external_idstring='extid',
                reason='because',
                sourcestamps=[234],
                submitted_at=datetime.datetime(1970, 4, 13, 21, 8, 8, tzinfo=UTC),
                results=-1,
            ),
        )

    @defer.inlineCallbacks
    def do_test_getBuildsetProperties(self, buildsetid, rows, expected):
        yield self.db.insert_test_data(rows)
        props = yield self.db.buildsets.getBuildsetProperties(buildsetid)

        self.assertEqual(props, expected)

    def test_getBuildsetProperties_multiple(self):
        return self.do_test_getBuildsetProperties(
            91,
            [
                fakedb.Buildset(id=91, complete=0, results=-1, submitted_at=0),
                fakedb.BuildsetProperty(
                    buildsetid=91, property_name='prop1', property_value='["one", "fake1"]'
                ),
                fakedb.BuildsetProperty(
                    buildsetid=91, property_name='prop2', property_value='["two", "fake2"]'
                ),
            ],
            {"prop1": ('one', 'fake1'), "prop2": ('two', 'fake2')},
        )

    def test_getBuildsetProperties_empty(self):
        return self.do_test_getBuildsetProperties(
            91,
            [
                fakedb.Buildset(id=91, complete=0, results=-1, submitted_at=0),
            ],
            {},
        )

    def test_getBuildsetProperties_nosuch(self):
        "returns an empty dict even if no such buildset exists"
        return self.do_test_getBuildsetProperties(91, [], {})

    @defer.inlineCallbacks
    def test_getBuildset_incomplete_zero(self):
        yield self.db.insert_test_data([
            fakedb.Buildset(
                id=91,
                complete=0,
                complete_at=0,
                results=-1,
                submitted_at=266761875,
                external_idstring='extid',
                reason='rsn',
            ),
            fakedb.BuildsetSourceStamp(buildsetid=91, sourcestampid=234),
        ])
        bsdict = yield self.db.buildsets.getBuildset(91)

        self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            bsdict,
            buildsets.BuildSetModel(
                bsid=91,
                external_idstring='extid',
                reason='rsn',
                sourcestamps=[234],
                submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 15, tzinfo=UTC),
                complete=False,
                complete_at=epoch2datetime(0),
                results=-1,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuildset_complete(self):
        yield self.db.insert_test_data([
            fakedb.Buildset(
                id=91,
                complete=1,
                complete_at=298297875,
                results=-1,
                submitted_at=266761875,
                external_idstring='extid',
                reason='rsn',
            ),
            fakedb.BuildsetSourceStamp(buildsetid=91, sourcestampid=234),
        ])
        bsdict = yield self.db.buildsets.getBuildset(91)

        self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            bsdict,
            buildsets.BuildSetModel(
                bsid=91,
                external_idstring='extid',
                reason='rsn',
                sourcestamps=[234],
                submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 15, tzinfo=UTC),
                complete=True,
                complete_at=datetime.datetime(1979, 6, 15, 12, 31, 15, tzinfo=UTC),
                results=-1,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuildset_nosuch(self):
        bsdict = yield self.db.buildsets.getBuildset(91)

        self.assertEqual(bsdict, None)

    def insert_test_getBuildsets_data(self):
        return self.db.insert_test_data([
            fakedb.Buildset(
                id=91,
                complete=0,
                complete_at=298297875,
                results=-1,
                submitted_at=266761875,
                external_idstring='extid',
                reason='rsn1',
            ),
            fakedb.BuildsetSourceStamp(buildsetid=91, sourcestampid=234),
            fakedb.Buildset(
                id=92,
                complete=1,
                complete_at=298297876,
                results=7,
                submitted_at=266761876,
                external_idstring='extid',
                reason='rsn2',
            ),
            fakedb.BuildsetSourceStamp(buildsetid=92, sourcestampid=234),
        ])

    @defer.inlineCallbacks
    def test_getBuildsets_empty(self):
        bsdictlist = yield self.db.buildsets.getBuildsets()

        self.assertEqual(bsdictlist, [])

    @defer.inlineCallbacks
    def test_getBuildsets_all(self):
        yield self.insert_test_getBuildsets_data()
        bsdictlist = yield self.db.buildsets.getBuildsets()

        def bsdictKey(bsdict):
            return bsdict.reason

        for bsdict in bsdictlist:
            self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            sorted(bsdictlist, key=bsdictKey),
            sorted(
                [
                    buildsets.BuildSetModel(
                        bsid=91,
                        external_idstring='extid',
                        reason='rsn1',
                        sourcestamps=[234],
                        submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 15, tzinfo=UTC),
                        complete_at=datetime.datetime(1979, 6, 15, 12, 31, 15, tzinfo=UTC),
                        complete=False,
                        results=-1,
                    ),
                    buildsets.BuildSetModel(
                        bsid=92,
                        external_idstring='extid',
                        reason='rsn2',
                        sourcestamps=[234],
                        submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 16, tzinfo=UTC),
                        complete_at=datetime.datetime(1979, 6, 15, 12, 31, 16, tzinfo=UTC),
                        complete=True,
                        results=7,
                    ),
                ],
                key=bsdictKey,
            ),
        )

    @defer.inlineCallbacks
    def test_getBuildsets_complete(self):
        yield self.insert_test_getBuildsets_data()
        bsdictlist = yield self.db.buildsets.getBuildsets(complete=True)

        for bsdict in bsdictlist:
            self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            bsdictlist,
            [
                buildsets.BuildSetModel(
                    bsid=92,
                    external_idstring='extid',
                    reason='rsn2',
                    sourcestamps=[234],
                    submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 16, tzinfo=UTC),
                    complete_at=datetime.datetime(1979, 6, 15, 12, 31, 16, tzinfo=UTC),
                    complete=True,
                    results=7,
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_getBuildsets_incomplete(self):
        yield self.insert_test_getBuildsets_data()
        bsdictlist = yield self.db.buildsets.getBuildsets(complete=False)

        for bsdict in bsdictlist:
            self.assertIsInstance(bsdict, buildsets.BuildSetModel)
        self.assertEqual(
            bsdictlist,
            [
                buildsets.BuildSetModel(
                    bsid=91,
                    external_idstring='extid',
                    reason='rsn1',
                    sourcestamps=[234],
                    submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 15, tzinfo=UTC),
                    complete_at=datetime.datetime(1979, 6, 15, 12, 31, 15, tzinfo=UTC),
                    complete=False,
                    results=-1,
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_completeBuildset_already_completed(self):
        yield self.insert_test_getBuildsets_data()
        with self.assertRaises(buildsets.AlreadyCompleteError):
            yield self.db.buildsets.completeBuildset(bsid=92, results=6)

    @defer.inlineCallbacks
    def test_completeBuildset_missing(self):
        yield self.insert_test_getBuildsets_data()
        with self.assertRaises(buildsets.AlreadyCompleteError):
            yield self.db.buildsets.completeBuildset(bsid=93, results=6)

    @defer.inlineCallbacks
    def test_completeBuildset(self):
        yield self.insert_test_getBuildsets_data()
        yield self.db.buildsets.completeBuildset(bsid=91, results=6)
        bsdicts = yield self.db.buildsets.getBuildsets()

        bsdicts = [
            (
                bsdict.bsid,
                bsdict.complete,
                datetime2epoch(bsdict.complete_at),
                bsdict.results,
            )
            for bsdict in bsdicts
        ]
        self.assertEqual(sorted(bsdicts), sorted([(91, 1, self.now, 6), (92, 1, 298297876, 7)]))

    @defer.inlineCallbacks
    def test_completeBuildset_explicit_complete_at(self):
        yield self.insert_test_getBuildsets_data()
        yield self.db.buildsets.completeBuildset(
            bsid=91, results=6, complete_at=epoch2datetime(72759)
        )
        bsdicts = yield self.db.buildsets.getBuildsets()

        bsdicts = [
            (
                bsdict.bsid,
                bsdict.complete,
                datetime2epoch(bsdict.complete_at),
                bsdict.results,
            )
            for bsdict in bsdicts
        ]
        self.assertEqual(sorted(bsdicts), sorted([(91, 1, 72759, 6), (92, 1, 298297876, 7)]))

    def insert_test_getRecentBuildsets_data(self):
        return self.db.insert_test_data([
            fakedb.SourceStamp(id=91, branch='branch_a', repository='repo_a'),
            fakedb.Buildset(
                id=91,
                complete=0,
                complete_at=298297875,
                results=-1,
                submitted_at=266761875,
                external_idstring='extid',
                reason='rsn1',
            ),
            fakedb.BuildsetSourceStamp(buildsetid=91, sourcestampid=91),
            fakedb.Buildset(
                id=92,
                complete=1,
                complete_at=298297876,
                results=7,
                submitted_at=266761876,
                external_idstring='extid',
                reason='rsn2',
            ),
            fakedb.BuildsetSourceStamp(buildsetid=92, sourcestampid=91),
            # buildset unrelated to the change
            fakedb.Buildset(
                id=93,
                complete=1,
                complete_at=298297877,
                results=7,
                submitted_at=266761877,
                external_idstring='extid',
                reason='rsn2',
            ),
        ])

    @defer.inlineCallbacks
    def test_getRecentBuildsets_all(self):
        yield self.insert_test_getRecentBuildsets_data()
        bsdictlist = yield self.db.buildsets.getRecentBuildsets(
            2, branch='branch_a', repository='repo_a'
        )

        self.assertEqual(
            bsdictlist,
            [
                buildsets.BuildSetModel(
                    bsid=91,
                    external_idstring='extid',
                    reason='rsn1',
                    sourcestamps=[91],
                    submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 15, tzinfo=UTC),
                    complete_at=datetime.datetime(1979, 6, 15, 12, 31, 15, tzinfo=UTC),
                    complete=False,
                    results=-1,
                ),
                buildsets.BuildSetModel(
                    bsid=92,
                    external_idstring='extid',
                    reason='rsn2',
                    sourcestamps=[91],
                    submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 16, tzinfo=UTC),
                    complete_at=datetime.datetime(1979, 6, 15, 12, 31, 16, tzinfo=UTC),
                    complete=True,
                    results=7,
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_getRecentBuildsets_one(self):
        yield self.insert_test_getRecentBuildsets_data()
        bsdictlist = yield self.db.buildsets.getRecentBuildsets(
            1, branch='branch_a', repository='repo_a'
        )

        self.assertEqual(
            bsdictlist,
            [
                buildsets.BuildSetModel(
                    bsid=92,
                    external_idstring='extid',
                    reason='rsn2',
                    sourcestamps=[91],
                    submitted_at=datetime.datetime(1978, 6, 15, 12, 31, 16, tzinfo=UTC),
                    complete_at=datetime.datetime(1979, 6, 15, 12, 31, 16, tzinfo=UTC),
                    complete=True,
                    results=7,
                ),
            ],
        )

    @defer.inlineCallbacks
    def test_getRecentBuildsets_zero(self):
        yield self.insert_test_getRecentBuildsets_data()
        bsdictlist = yield self.db.buildsets.getRecentBuildsets(
            0, branch='branch_a', repository='repo_a'
        )

        self.assertEqual(bsdictlist, [])

    @defer.inlineCallbacks
    def test_getRecentBuildsets_noBranchMatch(self):
        yield self.insert_test_getRecentBuildsets_data()
        bsdictlist = yield self.db.buildsets.getRecentBuildsets(
            2, branch='bad_branch', repository='repo_a'
        )

        self.assertEqual(bsdictlist, [])

    @defer.inlineCallbacks
    def test_getRecentBuildsets_noRepoMatch(self):
        yield self.insert_test_getRecentBuildsets_data()
        bsdictlist = yield self.db.buildsets.getRecentBuildsets(
            2, branch='branch_a', repository='bad_repo'
        )

        self.assertEqual(bsdictlist, [])

    @defer.inlineCallbacks
    def test_addBuildset_simple(self):
        (bsid, brids) = yield self.db.buildsets.addBuildset(
            sourcestamps=[234],
            reason='because',
            properties={},
            builderids=[2],
            external_idstring='extid',
            waited_for=True,
        )

        def thd(conn):
            # we should only have one brid
            self.assertEqual(len(brids), 1)

            # should see one buildset row
            r = conn.execute(self.db.model.buildsets.select())
            rows = [
                (
                    row.id,
                    row.external_idstring,
                    row.reason,
                    row.complete,
                    row.complete_at,
                    row.submitted_at,
                    row.results,
                )
                for row in r.fetchall()
            ]
            self.assertEqual(rows, [(bsid, 'extid', 'because', 0, None, self.now, -1)])

            # one buildrequests row
            r = conn.execute(self.db.model.buildrequests.select())
            self.assertEqual(
                r.keys(),
                [
                    'id',
                    'buildsetid',
                    'builderid',
                    'priority',
                    'complete',
                    'results',
                    'submitted_at',
                    'complete_at',
                    'waited_for',
                ],
            )
            self.assertEqual(r.fetchall(), [(bsid, brids[2], 2, 0, 0, -1, self.now, None, 1)])

            # one buildset_sourcestamps row
            r = conn.execute(self.db.model.buildset_sourcestamps.select())
            self.assertEqual(list(r.keys()), ['id', 'buildsetid', 'sourcestampid'])
            self.assertEqual(r.fetchall(), [(1, bsid, 234)])

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_addBuildset_bigger(self):
        props = {"prop": (['list'], 'test')}
        yield defer.succeed(None)
        xxx_todo_changeme1 = yield self.db.buildsets.addBuildset(
            sourcestamps=[234],
            reason='because',
            waited_for=False,
            properties=props,
            builderids=[1, 2],
        )

        (bsid, brids) = xxx_todo_changeme1

        def thd(conn):
            self.assertEqual(len(brids), 2)

            # should see one buildset row
            r = conn.execute(self.db.model.buildsets.select())
            rows = [
                (
                    row.id,
                    row.external_idstring,
                    row.reason,
                    row.complete,
                    row.complete_at,
                    row.results,
                )
                for row in r.fetchall()
            ]
            self.assertEqual(rows, [(bsid, None, 'because', 0, None, -1)])

            # one property row
            r = conn.execute(self.db.model.buildset_properties.select())
            rows = [(row.buildsetid, row.property_name, row.property_value) for row in r.fetchall()]
            self.assertEqual(rows, [(bsid, 'prop', json.dumps([['list'], 'test']))])

            # one buildset_sourcestamps row
            r = conn.execute(self.db.model.buildset_sourcestamps.select())
            rows = [(row.buildsetid, row.sourcestampid) for row in r.fetchall()]
            self.assertEqual(rows, [(bsid, 234)])

            # and two buildrequests rows (and don't re-check the default
            # columns)
            r = conn.execute(self.db.model.buildrequests.select())
            rows = [(row.buildsetid, row.id, row.builderid) for row in r.fetchall()]

            # we don't know which of the brids is assigned to which
            # buildername, but either one will do
            self.assertEqual(sorted(rows), [(bsid, brids[1], 1), (bsid, brids[2], 2)])

        yield self.db.pool.do(thd)

    @defer.inlineCallbacks
    def test_addBuildset_properties_cache(self):
        """
        Test that `addChange` properly seeds the `getChange` cache.
        """

        # Patchup the buildset properties cache so we can verify that
        # it got called form `addBuildset`.
        mockedCachePut = mock.Mock()
        self.patch(self.db.buildsets.getBuildsetProperties.cache, "put", mockedCachePut)

        # Setup a dummy set of properties to insert with the buildset.
        props = {"prop": (['list'], 'test')}

        # Now, call `addBuildset`, and verify that the above properties
        # were seed in the `getBuildsetProperties` cache.
        bsid, _ = yield self.db.buildsets.addBuildset(
            sourcestamps=[234],
            reason='because',
            properties=props,
            builderids=[1, 2],
            waited_for=False,
        )
        mockedCachePut.assert_called_once_with(bsid, props)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.trial import unittest

from buildbot.test.reactor import TestReactorMixin
from buildbot.util.watchdog import Watchdog


class TestWatchdog(TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()

    def test_not_started_no_calls(self):
        m = mock.Mock()
        w = Watchdog(self.reactor, m, 10)
        self.reactor.pump([1] * 100)
        self.assertEqual(m.call_count, 0)
        del w  # to silence unused variable warnings

    def test_started_calls(self):
        m = mock.Mock()
        w = Watchdog(self.reactor, m, 10)
        w.start()

        self.reactor.advance(9.9)
        self.assertEqual(m.call_count, 0)
        self.reactor.advance(0.2)
        self.assertEqual(m.call_count, 1)
        self.reactor.advance(20)
        self.assertEqual(m.call_count, 1)

    def test_two_starts_single_call(self):
        m = mock.Mock()
        w = Watchdog(self.reactor, m, 10)
        w.start()
        w.start()

        self.reactor.advance(9.9)
        self.assertEqual(m.call_count, 0)
        self.reactor.advance(0.2)
        self.assertEqual(m.call_count, 1)
        self.reactor.advance(20)
        self.assertEqual(m.call_count, 1)

    def test_started_stopped_does_not_call(self):
        m = mock.Mock()
        w = Watchdog(self.reactor, m, 10)
        w.start()
        w.stop()

        self.reactor.pump([1] * 100)
        self.assertEqual(m.call_count, 0)

    def test_triggers_repeatedly(self):
        m = mock.Mock()
        w = Watchdog(self.reactor, m, 10)

        w.start()
        self.reactor.advance(9.9)
        self.assertEqual(m.call_count, 0)
        self.reactor.advance(0.2)
        self.assertEqual(m.call_count, 1)

        w.start()
        self.reactor.advance(9.9)
        self.assertEqual(m.call_count, 1)
        self.reactor.advance(0.2)
        self.assertEqual(m.call_count, 2)

        w.start()
        self.reactor.advance(9.9)
        self.assertEqual(m.call_count, 2)
        self.reactor.advance(0.2)
        self.assertEqual(m.call_count, 3)

    def test_notify_delays_trigger(self):
        m = mock.Mock()
        w = Watchdog(self.reactor, m, 10)

        w.start()
        self.reactor.advance(5)
        w.notify()
        self.reactor.advance(9.9)
        self.assertEqual(m.call_count, 0)
        self.reactor.advance(0.2)
        self.assertEqual(m.call_count, 1)



#!/usr/bin/env python


import sys

from coverage import coverage
from coverage.results import Numbers
from coverage.summary import SummaryReporter
from twisted.python import usage

# this is an adaptation of the code behind "coverage report", modified to
# display+sortby "lines uncovered", which (IMHO) is more important of a
# metric than lines covered or percentage covered. Concentrating on the files
# with the most uncovered lines encourages getting the tree and test suite
# into a state that provides full line-coverage on all files.

# much of this code was adapted from coverage/summary.py in the 'coverage'
# distribution, and is used under their BSD license.


class Options(usage.Options):
    optParameters = [
        ("sortby", "s", "uncovered", "how to sort: uncovered, covered, name"),
    ]


class MyReporter(SummaryReporter):
    def report(self, outfile=None, sortby="uncovered"):
        self.find_code_units(
            None, ["/System", "/Library", "/usr/lib", "buildbot/test", "simplejson"]
        )

        # Prepare the formatting strings
        max_name = max([len(cu.name) for cu in self.code_units] + [5])
        fmt_name = f"%- {max_name}s  "
        fmt_err = "%s   %s: %s\n"
        header1 = (fmt_name % "") + "     Statements    "
        header2 = (fmt_name % "Name") + " Uncovered  Covered"
        fmt_coverage = fmt_name + "%9d  %7d "
        if self.branches:
            header1 += "   Branches   "
            header2 += " Found  Excutd"
            fmt_coverage += " %6d %6d"
        header1 += "  Percent"
        header2 += "  Covered"
        fmt_coverage += " %7d%%"
        if self.show_missing:
            header1 += "          "
            header2 += "   Missing"
            fmt_coverage += "   %s"
        rule = "-" * len(header1) + "\n"
        header1 += "\n"
        header2 += "\n"
        fmt_coverage += "\n"

        if not outfile:
            outfile = sys.stdout

        # Write the header
        outfile.write(header1)
        outfile.write(header2)
        outfile.write(rule)

        total = Numbers()
        total_uncovered = 0

        lines = []
        for cu in self.code_units:
            try:
                analysis = self.coverage._analyze(cu)
                nums = analysis.numbers
                uncovered = nums.n_statements - nums.n_executed
                total_uncovered += uncovered
                args = (cu.name, uncovered, nums.n_executed)
                if self.branches:
                    args += (nums.n_branches, nums.n_executed_branches)
                args += (nums.pc_covered,)
                if self.show_missing:
                    args += (analysis.missing_formatted(),)
                if sortby == "covered":
                    sortkey = nums.pc_covered
                elif sortby == "uncovered":
                    sortkey = uncovered
                else:
                    sortkey = cu.name
                lines.append((sortkey, fmt_coverage % args))
                total += nums
            except Exception:
                if not self.ignore_errors:
                    typ, msg = sys.exc_info()[:2]
                    outfile.write(fmt_err % (cu.name, typ.__name__, msg))
        lines.sort()
        if sortby in ("uncovered", "covered"):
            lines.reverse()
        for sortkey, line in lines:
            outfile.write(line)

        if total.n_files > 1:
            outfile.write(rule)
            args = ("TOTAL", total_uncovered, total.n_executed)
            if self.branches:
                args += (total.n_branches, total.n_executed_branches)
            args += (total.pc_covered,)
            if self.show_missing:
                args += ("",)
            outfile.write(fmt_coverage % args)


def report(o):
    c = coverage()
    c.load()
    r = MyReporter(c, show_missing=False, ignore_errors=False)
    r.report(sortby=o['sortby'])


if __name__ == '__main__':
    o = Options()
    o.parseOptions()
    report(o)



# Copyright Buildbot Team Members
# Portions copyright 2015-2016 ClusterHQ Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Sequence

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.base import _ThreePhaseEvent
from twisted.internet.error import ProcessDone
from twisted.internet.interfaces import IReactorCore
from twisted.internet.interfaces import IReactorThreads
from twisted.internet.task import Clock
from twisted.python import log
from twisted.python.failure import Failure
from zope.interface import implementer

from buildbot.util.twisted import async_to_deferred

# The code here is based on the implementations in
# https://twistedmatrix.com/trac/ticket/8295
# https://twistedmatrix.com/trac/ticket/8296


@implementer(IReactorCore)
class CoreReactor:
    """
    Partial implementation of ``IReactorCore``.
    """

    def __init__(self):
        super().__init__()
        self._triggers = {}

    def addSystemEventTrigger(
        self, phase: str, eventType: str, callable: Callable[..., Any], *args, **kw
    ):
        event = self._triggers.setdefault(eventType, _ThreePhaseEvent())
        return eventType, event.addTrigger(phase, callable, *args, **kw)

    def removeSystemEventTrigger(self, triggerID):
        eventType, handle = triggerID
        event = self._triggers.setdefault(eventType, _ThreePhaseEvent())
        event.removeTrigger(handle)

    def fireSystemEvent(self, eventType):
        event = self._triggers.get(eventType)
        if event is not None:
            event.fireEvent()

    def callWhenRunning(self, callable: Callable[..., Any], *args, **kwargs):
        callable(*args, **kwargs)

    def resolve(self, name: str, timeout: Sequence[int]) -> defer.Deferred[str]:  # type: ignore
        # Twisted 24.11 changed the default value of timeout, thus to support
        # both versions typing is disabled
        raise NotImplementedError("resolve() is not implemented in this reactor")

    def stop(self) -> None:
        raise NotImplementedError("stop() is not implemented in this reactor")

    def running(self) -> bool:
        raise NotImplementedError("running() is not implemented in this reactor")

    def crash(self):
        raise NotImplementedError("crash() is not implemented in this reactor")

    def iterate(self, delay=None):
        raise NotImplementedError("iterate() is not implemented in this reactor")

    def run(self):
        raise NotImplementedError("run() is not implemented in this reactor")

    def runUntilCurrent(self):
        raise NotImplementedError("runUntilCurrent() is not implemented in this reactor")


class NonThreadPool:
    """
    A stand-in for ``twisted.python.threadpool.ThreadPool`` so that the
    majority of the test suite does not need to use multithreading.

    This implementation takes the function call which is meant to run in a
    thread pool and runs it synchronously in the calling thread.

    :ivar int calls: The number of calls which have been dispatched to this
        object.
    """

    calls = 0

    def __init__(self, **kwargs):
        pass

    def callInThreadWithCallback(self, onResult, func, *args, **kw):
        self.calls += 1
        try:
            result = func(*args, **kw)
        except:  # noqa: E722
            # We catch *everything* here, since normally this code would be
            # running in a thread, where there is nothing that will catch
            # error.
            onResult(False, Failure())
        else:
            onResult(True, result)

    def start(self):
        pass

    def stop(self):
        pass


@implementer(IReactorThreads)
class NonReactor:
    """
    A partial implementation of ``IReactorThreads`` which fits into
    the execution model defined by ``NonThreadPool``.
    """

    def suggestThreadPoolSize(self, size: int) -> None:
        pass

    def callInThread(self, callable: Callable[..., Any], *args, **kwargs: object) -> None:
        callable(*args, **kwargs)

    def callFromThread(self, callable: Callable[..., Any], *args: object, **kwargs: object) -> None:  # type: ignore[override]
        callable(*args, **kwargs)

    def getThreadPool(self):
        return NonThreadPool()


class ProcessTransport:
    def __init__(self, pid, reactor):
        self.pid = pid
        self.reactor = reactor

    def signalProcess(self, signal):
        self.reactor.process_signalProcess(self.pid, signal)

    def closeChildFD(self, descriptor):
        self.reactor.process_closeChildFD(self.pid, descriptor)


class ProcessReactor:
    def __init__(self):
        super().__init__()
        self._expected_calls = []
        self._processes = {}
        self._next_pid = 0

    def _check_call(self, call):
        function = call["call"]
        if not self._expected_calls:
            raise AssertionError(f"Unexpected call to {function}(): {call!r}")
        expected_call = self._expected_calls.pop(0)
        self.testcase.assertEqual(call, expected_call, f"when calling to {function}()")

    def spawnProcess(
        self,
        processProtocol,
        executable,
        args,
        env=None,
        path=None,
        uid=None,
        gid=None,
        usePTY=False,
        childFDs=None,
    ):
        call = {
            "call": "spawnProcess",
            "executable": executable,
            "args": args,
            "env": env,
            "path": path,
            "uid": uid,
            "gid": gid,
            "usePTY": usePTY,
            "childFDs": childFDs,
        }
        self._check_call(call)

        pid = self._next_pid
        self._next_pid += 1

        self._processes[pid] = processProtocol
        return ProcessTransport(pid, self)

    def process_signalProcess(self, pid, signal):
        call = {
            "call": "signalProcess",
            "pid": pid,
            "signal": signal,
        }
        self._check_call(call)

    def process_closeChildFD(self, pid, descriptor):
        call = {
            "call": "closeChildFD",
            "pid": pid,
            "descriptor": descriptor,
        }
        self._check_call(call)

    def process_done(self, pid, status):
        reason = ProcessDone(status)
        self._processes[pid].processEnded(reason)
        self._processes[pid].processExited(reason)

    def process_send_stdout(self, pid, data):
        self._processes[pid].outReceived(data)

    def process_send_stderr(self, pid, data):
        self._processes[pid].errReceived(data)

    def assert_no_remaining_calls(self):
        if self._expected_calls:
            msg = "The following expected calls are missing: "
            for call in self._expected_calls:
                copy = call.copy()
                name = copy.pop("call")
                msg += f"\nTo {name}(): {call!r}"
            raise AssertionError(msg)

    def expect_spawn(
        self, executable, args, env=None, path=None, uid=None, gid=None, usePTY=False, childFDs=None
    ):
        self._expected_calls.append({
            "call": "spawnProcess",
            "executable": executable,
            "args": args,
            "env": env,
            "path": path,
            "uid": uid,
            "gid": gid,
            "usePTY": usePTY,
            "childFDs": childFDs,
        })

    def expect_process_signalProcess(self, pid, signal):
        self._expected_calls.append({
            "call": "signalProcess",
            "pid": pid,
            "signal": signal,
        })

    def expect_process_closeChildFD(self, pid, descriptor):
        self._expected_calls.append({
            "call": "closeChildFD",
            "pid": pid,
            "descriptor": descriptor,
        })


class TestReactor(ProcessReactor, NonReactor, CoreReactor, Clock):
    def __init__(self):
        super().__init__()

        # whether there are calls that should run right now
        self._pendingCurrentCalls = False
        self.stop_called = False

    def set_test_case(self, testcase):
        self.testcase = testcase

    def _executeCurrentDelayedCalls(self):
        while self.getDelayedCalls():
            first = sorted(self.getDelayedCalls(), key=lambda a: a.getTime())[0]
            if first.getTime() > self.seconds():
                break
            self.advance(0)

        self._pendingCurrentCalls = False

    @async_to_deferred
    async def _catchPrintExceptions(self, what, *a, **kw) -> None:
        try:
            await defer.maybeDeferred(what, *a, **kw)
        except Exception as e:
            log.msg('Unhandled exception from deferred when doing TestReactor.advance()', e)
            raise

    def callLater(self, when, what, *a, **kw):
        # Buildbot often uses callLater(0, ...) to defer execution of certain
        # code to the next iteration of the reactor. This means that often
        # there are pending callbacks registered to the reactor that might
        # block other code from proceeding unless the test reactor has an
        # iteration. To avoid deadlocks in tests we give the real reactor a
        # chance to advance the test reactor whenever we detect that there
        # are callbacks that should run in the next iteration of the test
        # reactor.
        #
        # Additionally, we wrap all calls with a function that prints any
        # unhandled exceptions
        if when <= 0 and not self._pendingCurrentCalls:
            reactor.callLater(0, self._executeCurrentDelayedCalls)

        return super().callLater(when, self._catchPrintExceptions, what, *a, **kw)

    def stop(self):
        # first fire pending calls until the current time. Note that the real
        # reactor only advances until the current time in the case of shutdown.
        self.advance(0)

        # then, fire the shutdown event
        self.fireSystemEvent('shutdown')

        self.stop_called = True



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from twisted.internet import defer
from zope.interface import implementer

from buildbot.config import BuilderConfig
from buildbot.interfaces import IBuildStepFactory
from buildbot.machine.base import Machine
from buildbot.process.buildstep import BuildStep
from buildbot.process.buildstep import create_step_from_step_or_factory
from buildbot.process.factory import BuildFactory
from buildbot.process.properties import Interpolate
from buildbot.process.results import CANCELLED
from buildbot.process.results import RETRY
from buildbot.test.fake.latent import LatentController
from buildbot.test.fake.step import BuildStepController
from buildbot.test.fake.worker import WorkerController
from buildbot.test.util.integration import RunFakeMasterTestCase

RemoteWorker: type | None = None
try:
    from buildbot_worker.bot import LocalWorker as RemoteWorker
except ImportError:
    pass


@implementer(IBuildStepFactory)
class StepController:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.steps = []

    def buildStep(self):
        step_deferred = defer.Deferred()
        step = create_step_from_step_or_factory(ControllableStep(step_deferred, **self.kwargs))
        self.steps.append((step, step_deferred))
        return step


class ControllableStep(BuildStep):
    def run(self):
        return self._step_deferred

    def __init__(self, step_deferred, **kwargs):
        super().__init__(**kwargs)
        self._step_deferred = step_deferred

    def interrupt(self, reason):
        self._step_deferred.callback(CANCELLED)


class Tests(RunFakeMasterTestCase):
    @defer.inlineCallbacks
    def test_latent_max_builds(self):
        """
        If max_builds is set, only one build is started on a latent
        worker at a time.
        """
        controller = LatentController(self, 'local', max_builds=1)
        step_controller = StepController()
        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy-1",
                    workernames=["local"],
                    factory=BuildFactory([step_controller]),
                ),
                BuilderConfig(
                    name="testy-2",
                    workernames=["local"],
                    factory=BuildFactory([step_controller]),
                ),
            ],
            'workers': [controller.worker],
            'protocols': {'null': {}},
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_ids = [
            (yield self.master.data.updates.findBuilderId('testy-1')),
            (yield self.master.data.updates.findBuilderId('testy-2')),
        ]

        started_builds = []
        yield self.master.mq.startConsuming(
            lambda key, build: started_builds.append(build), ('builds', None, 'new')
        )

        # Trigger a buildrequest
        yield self.master.data.updates.addBuildset(
            waited_for=False,
            builderids=builder_ids,
            sourcestamps=[
                {'codebase': '', 'repository': '', 'branch': None, 'revision': None, 'project': ''},
            ],
        )
        self.master.reactor.advance(1)

        # The worker fails to substantiate.
        controller.start_instance(True)

        yield controller.connect_worker()

        self.assertEqual(len(started_builds), 1)
        yield controller.auto_stop(True)

    @defer.inlineCallbacks
    def test_local_worker_max_builds(self):
        """
        If max_builds is set, only one build is started on a worker
        at a time.
        """
        step_controller = StepController()
        config_dict = {
            'builders': [
                BuilderConfig(
                    name="testy-1",
                    workernames=["local"],
                    factory=BuildFactory([step_controller]),
                ),
                BuilderConfig(
                    name="testy-2",
                    workernames=["local"],
                    factory=BuildFactory([step_controller]),
                ),
            ],
            'workers': [self.createLocalWorker('local', max_builds=1)],
            'protocols': {'null': {}},
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)
        builder_ids = [
            (yield self.master.data.updates.findBuilderId('testy-1')),
            (yield self.master.data.updates.findBuilderId('testy-2')),
        ]

        started_builds = []
        yield self.master.mq.startConsuming(
            lambda key, build: started_builds.append(build), ('builds', None, 'new')
        )

        # Trigger a buildrequest
        yield self.master.data.updates.addBuildset(
            waited_for=False,
            builderids=builder_ids,
            sourcestamps=[
                {'codebase': '', 'repository': '', 'branch': None, 'revision': None, 'project': ''},
            ],
        )
        self.master.reactor.advance(1)

        self.assertEqual(len(started_builds), 1)

    @defer.inlineCallbacks
    def test_worker_registered_to_machine(self):
        worker = self.createLocalWorker('worker1', machine_name='machine1')
        machine = Machine('machine1')

        config_dict = {
            'builders': [
                BuilderConfig(
                    name="builder1",
                    workernames=["worker1"],
                    factory=BuildFactory(),
                ),
            ],
            'workers': [worker],
            'machines': [machine],
            'protocols': {'null': {}},
            'multiMaster': True,
        }

        yield self.setup_master(config_dict)

        self.assertIs(worker.machine, machine)

    @defer.inlineCallbacks
    def test_worker_reconfigure_with_new_builder(self):
        """
        Checks if we can successfully reconfigure if we add new builders to worker.
        """
        config_dict = {
            'builders': [
                BuilderConfig(name="builder1", workernames=['local1'], factory=BuildFactory()),
            ],
            'workers': [self.createLocalWorker('local1', max_builds=1)],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)

        config_dict['builders'] += [
            BuilderConfig(name="builder2", workernames=['local1'], factory=BuildFactory()),
        ]
        config_dict['workers'] = [self.createLocalWorker('local1', max_builds=2)]

        # reconfig should succeed
        yield self.reconfig_master(config_dict)

    @defer.inlineCallbacks
    def test_step_with_worker_build_props_during_worker_disconnect(self):
        """
        We need to handle worker disconnection and steps with worker build properties gracefully
        """
        controller = WorkerController(self, 'local')

        stepcontroller = BuildStepController()

        config_dict = {
            'builders': [
                BuilderConfig(
                    name="builder",
                    workernames=['local'],
                    properties={'worker': Interpolate("%(worker:numcpus)s")},
                    factory=BuildFactory([stepcontroller.step]),
                ),
            ],
            'workers': [controller.worker],
            'protocols': {'null': {}},
            'multiMaster': True,
        }

        yield self.setup_master(config_dict)

        builder_id = yield self.master.data.updates.findBuilderId('builder')
        yield self.create_build_request([builder_id])

        yield controller.connect_worker()
        self.reactor.advance(1)
        yield controller.disconnect_worker()
        self.reactor.advance(1)
        yield self.assertBuildResults(1, RETRY)

    @defer.inlineCallbacks
    def test_worker_os_release_info_roundtrip(self):
        """
        Checks if we can successfully get information about the platform the worker is running on.
        This is very similar to test_worker_comm.TestWorkerComm.test_worker_info, except that
        we check details such as whether the information is passed in correct encoding.
        """
        worker = self.createLocalWorker('local1')

        config_dict = {
            'builders': [
                BuilderConfig(name="builder1", workernames=['local1'], factory=BuildFactory()),
            ],
            'workers': [worker],
            'protocols': {'null': {}},
            # Disable checks about missing scheduler.
            'multiMaster': True,
        }
        yield self.setup_master(config_dict)

        props = worker.info

        from buildbot_worker.base import BotBase

        expected_props_dict = {}
        BotBase._read_os_release(BotBase.os_release_file, expected_props_dict)

        for key, value in expected_props_dict.items():
            self.assertTrue(isinstance(key, str))
            self.assertTrue(isinstance(value, str))
            self.assertEqual(props.getProperty(key), value)

    if RemoteWorker is None:
        skip = "buildbot-worker package is not installed"



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""
Helpers for handling compatibility differences
between Python 2 and Python 3.
"""

from __future__ import annotations

from io import StringIO as NativeStringIO
from typing import TYPE_CHECKING
from typing import overload

if TYPE_CHECKING:
    from typing import Any
    from typing import TypeVar

    _T = TypeVar('_T')


@overload
def bytes2NativeString(x: bytes, encoding: str = 'utf-8') -> str: ...


@overload
def bytes2NativeString(x: _T, encoding: str = 'utf-8') -> _T: ...


def bytes2NativeString(x: _T, encoding: str = 'utf-8') -> str | _T:
    """
    Convert C{bytes} to a native C{str}.

    On Python 3 and higher, str and bytes
    are not equivalent.  In this case, decode
    the bytes, and return a native string.

    On Python 2 and lower, str and bytes
    are equivalent.  In this case, just
    just return the native string.

    @param x: a string of type C{bytes}
    @param encoding: an optional codec, default: 'utf-8'
    @return: a string of type C{str}
    """
    if isinstance(x, bytes) and str != bytes:
        return x.decode(encoding)
    return x


@overload
def unicode2bytes(x: str, encoding: str = 'utf-8', errors: str = 'strict') -> bytes: ...


@overload
def unicode2bytes(x: _T, encoding: str = 'utf-8', errors: str = 'strict') -> _T: ...


def unicode2bytes(x: str | _T, encoding: str = 'utf-8', errors: str = 'strict') -> bytes | _T:
    """
    Convert a unicode string to C{bytes}.

    @param x: a unicode string, of type C{str}.
    @param encoding: an optional codec, default: 'utf-8'
    @param errors: error handling scheme, default 'strict'
    @return: a string of type C{bytes}
    """
    if isinstance(x, str):
        return x.encode(encoding, errors)
    return x


@overload
def bytes2unicode(x: None, encoding: str = 'utf-8', errors: str = 'strict') -> None: ...


@overload
def bytes2unicode(x: Any, encoding: str = 'utf-8', errors: str = 'strict') -> str: ...


def bytes2unicode(x: Any | None, encoding: str = 'utf-8', errors: str = 'strict') -> str | None:
    """
    Convert a C{bytes} to a unicode string.

    @param x: a unicode string, of type C{str}.
    @param encoding: an optional codec, default: 'utf-8'
    @param errors: error handling scheme, default 'strict'
    @return: a unicode string of type C{unicode} on Python 2, or
             C{str} on Python 3.
    """
    if x is None:
        return None
    if isinstance(x, str):
        return x
    return str(x, encoding, errors)


__all__ = ["NativeStringIO", "bytes2NativeString", "bytes2unicode", "unicode2bytes"]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import enum
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.python import log

if TYPE_CHECKING:
    from buildbot.process.builder import Builder
    from buildbot.worker.base import AbstractWorker
    from buildbot.worker.latent import AbstractLatentWorker


class States(enum.Enum):
    # The worker isn't attached, or is in the process of attaching.
    DETACHED = 0
    # The worker is available to build: either attached, or a latent worker.
    AVAILABLE = 1
    # The worker is building.
    BUILDING = 2


class AbstractWorkerForBuilder:
    def __init__(self, builder: Builder):
        self.ping_watchers: list[defer.Deferred] = []
        self.state: States | None = None  # set in subclass
        self.worker: AbstractWorker | None = None
        self.builder = builder
        self.builder_name = builder.name
        self.locks = None

    def __repr__(self):
        r = ["<", self.__class__.__name__]
        if self.builder_name:
            r.extend([" builder=", repr(self.builder_name)])
        if self.worker:
            r.extend([" worker=", repr(self.worker.workername)])
        r.extend([" state=", self.state.name, ">"])
        return ''.join(r)

    def getWorkerCommandVersion(self, command, oldversion=None):
        if self.remoteCommands is None:
            # the worker is 0.5.0 or earlier
            return oldversion
        return self.remoteCommands.get(command)

    def isAvailable(self):
        # if this WorkerForBuilder is busy, then it's definitely not available
        if self.isBusy():
            return False

        # otherwise, check in with the Worker
        if self.worker:
            return self.worker.canStartBuild()

        # no worker? not very available.
        return False

    def isBusy(self):
        return self.state != States.AVAILABLE

    def buildStarted(self):
        self.state = States.BUILDING
        self.worker.buildStarted(self)

    def buildFinished(self):
        self.state = States.AVAILABLE
        if self.worker:
            self.worker.buildFinished(self)

    @defer.inlineCallbacks
    def attached(self, worker: AbstractWorker, commands):
        """
        @type  worker: L{buildbot.worker.Worker}
        @param worker: the Worker that represents the worker as a whole
        @type  commands: dict: string -> string, or None
        @param commands: provides the worker's version of each RemoteCommand
        """
        self.remoteCommands = commands  # maps command name to version
        if self.worker is None:
            self.worker = worker
            self.worker.addWorkerForBuilder(self)
        else:
            assert self.worker == worker
        log.msg(f"Worker {worker.workername} attached to {self.builder_name}")

        yield self.worker.conn.remotePrint(message="attached")

    def substantiate_if_needed(self, build):
        return defer.succeed(True)

    def insubstantiate_if_needed(self):
        pass

    def ping(self):
        """Ping the worker to make sure it is still there. Returns a Deferred
        that fires with True if it is.
        """
        newping = not self.ping_watchers
        d = defer.Deferred()
        self.ping_watchers.append(d)
        if newping:
            Ping().ping(self.worker.conn).addBoth(self._pong)

        return d

    def abortPingIfAny(self):
        watchers = self.ping_watchers
        self.ping_watchers = []
        for d in watchers:
            d.errback(PingException('aborted ping'))

    def _pong(self, res):
        watchers = self.ping_watchers
        self.ping_watchers = []
        for d in watchers:
            d.callback(res)

    def detached(self):
        log.msg(f"Worker {self.worker.workername} detached from {self.builder_name}")
        if self.worker:
            self.worker.removeWorkerForBuilder(self)
        self.worker = None
        self.remoteCommands = None


class PingException(Exception):
    pass


class Ping:
    running = False

    def ping(self, conn):
        assert not self.running
        if not conn:
            # clearly the ping must fail
            return defer.fail(PingException("Worker not connected?"))
        self.running = True
        log.msg("sending ping")
        self.d = defer.Deferred()
        # TODO: add a distinct 'ping' command on the worker.. using 'print'
        # for this purpose is kind of silly.
        conn.remotePrint(message="ping").addCallbacks(
            self._pong, self._ping_failed, errbackArgs=(conn,)
        )
        return self.d

    def _pong(self, res):
        log.msg("ping finished: success")
        self.d.callback(True)

    def _ping_failed(self, res, conn):
        log.msg("ping finished: failure")
        # the worker has some sort of internal error, disconnect them. If we
        # don't, we'll requeue a build and ping them again right away,
        # creating a nasty loop.
        conn.loseConnection()
        self.d.errback(res)


class WorkerForBuilder(AbstractWorkerForBuilder):
    def __init__(self, builder: Builder):
        super().__init__(builder)
        self.state = States.DETACHED

    @defer.inlineCallbacks
    def attached(self, worker, commands):
        yield super().attached(worker, commands)

        # Only set available on non-latent workers, since latent workers
        # only attach while a build is in progress.
        self.state = States.AVAILABLE

    def detached(self):
        super().detached()
        if self.worker:
            self.worker.removeWorkerForBuilder(self)
        self.worker = None
        self.state = States.DETACHED


class LatentWorkerForBuilder(AbstractWorkerForBuilder):
    def __init__(self, worker: AbstractLatentWorker, builder: Builder):
        super().__init__(builder)
        self.worker: AbstractLatentWorker | None = worker
        self.state = States.AVAILABLE
        self.worker.addWorkerForBuilder(self)
        log.msg(f"Latent worker {worker.workername} attached to {self.builder_name}")

    def substantiate_if_needed(self, build):
        self.state = States.DETACHED
        d = self.substantiate(build)
        return d

    def insubstantiate_if_needed(self):
        if self.worker:
            self.worker.insubstantiate()

    def attached(self, worker, commands):
        # When a latent worker is attached, it is actually because it prepared for a build
        # thus building and not available like for normal worker
        if self.state == States.DETACHED:
            self.state = States.BUILDING
        return super().attached(worker, commands)

    def substantiate(self, build):
        if self.worker:
            return self.worker.substantiate(self, build)
        return defer.succeed(False)

    def ping(self):
        if not self.worker.substantiated:
            return defer.fail(PingException("worker is not substantiated"))
        return super().ping()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.properties import Properties
from buildbot.process.properties import Property
from buildbot.test.fake.fakebuild import FakeBuild
from buildbot.util.git_credential import GitCredentialInputRenderer


class TestGitCredentialInputRenderer(unittest.TestCase):
    def setUp(self):
        self.props = Properties()
        self.build = FakeBuild(props=self.props)

    @defer.inlineCallbacks
    def test_render(self):
        self.props.setProperty("password", "property_password", "test")
        renderer = GitCredentialInputRenderer(
            username="user",
            password=Property("password"),
            url="https://example.com/repo.git",
        )
        rendered = yield self.build.render(renderer)
        self.assertEqual(
            rendered,
            "url=https://example.com/repo.git\nusername=user\npassword=property_password\n",
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.internet import defer
from twisted.trial import unittest
from zope.interface import implementer

from buildbot import interfaces
from buildbot.process import botmaster
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import service
from buildbot.worker import manager as workermanager


@implementer(interfaces.IWorker)
class FakeWorker(service.BuildbotService):
    reconfig_count = 0

    def __init__(self, workername):
        super().__init__(name=workername)

    def reconfigService(self):
        self.reconfig_count += 1
        self.configured = True
        return defer.succeed(None)

    @property
    def workername(self):
        return self.name


class FakeWorker2(FakeWorker):
    pass


class TestWorkerManager(TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantData=True)
        self.master.mq = self.master.mq
        self.workers = workermanager.WorkerManager(self.master)
        yield self.workers.setServiceParent(self.master)
        # workers expect a botmaster as well as a manager.
        self.master.botmaster.disownServiceParent()
        self.botmaster = botmaster.BotMaster()
        self.master.botmaster = self.botmaster
        yield self.master.botmaster.setServiceParent(self.master)

        self.new_config = mock.Mock()
        self.workers.startService()
        self.addCleanup(self.workers.stopService)

    @defer.inlineCallbacks
    def test_reconfigServiceWorkers_add_remove(self):
        worker = FakeWorker('worker1')
        self.new_config.workers = [worker]

        yield self.workers.reconfigServiceWithBuildbotConfig(self.new_config)

        self.assertIdentical(worker.parent, self.workers)
        self.assertEqual(self.workers.workers, {'worker1': worker})

        self.new_config.workers = []

        self.assertEqual(worker.running, True)
        yield self.workers.reconfigServiceWithBuildbotConfig(self.new_config)

        self.assertEqual(worker.running, False)

    @defer.inlineCallbacks
    def test_reconfigServiceWorkers_reconfig(self):
        worker = FakeWorker('worker1')
        yield worker.setServiceParent(self.workers)
        worker.parent = self.master
        worker.manager = self.workers
        worker.botmaster = self.master.botmaster

        worker_new = FakeWorker('worker1')
        self.new_config.workers = [worker_new]

        yield self.workers.reconfigServiceWithBuildbotConfig(self.new_config)

        # worker was not replaced..
        self.assertIdentical(self.workers.workers['worker1'], worker)

    @defer.inlineCallbacks
    def test_reconfigServiceWorkers_class_changes(self):
        worker = FakeWorker('worker1')
        yield worker.setServiceParent(self.workers)

        worker_new = FakeWorker2('worker1')
        self.new_config.workers = [worker_new]

        yield self.workers.reconfigServiceWithBuildbotConfig(self.new_config)

        # worker *was* replaced (different class)
        self.assertIdentical(self.workers.workers['worker1'], worker_new)

    @defer.inlineCallbacks
    def test_newConnection_remoteGetWorkerInfo_failure(self):
        class Error(RuntimeError):
            pass

        conn = mock.Mock()
        conn.remoteGetWorkerInfo = mock.Mock(return_value=defer.fail(Error()))
        with self.assertRaises(Error):
            yield self.workers.newConnection(conn, "worker")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import test_result_sets
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.twisted import async_to_deferred


class Tests(TestReactorMixin, unittest.TestCase):
    common_data = [
        fakedb.Worker(id=47, name='linux'),
        fakedb.Buildset(id=20),
        fakedb.Builder(id=88, name='b1'),
        fakedb.Builder(id=89, name='b2'),
        fakedb.BuildRequest(id=41, buildsetid=20, builderid=88),
        fakedb.BuildRequest(id=42, buildsetid=20, builderid=88),
        fakedb.BuildRequest(id=43, buildsetid=20, builderid=88),
        fakedb.Master(id=88),
        fakedb.Build(id=30, buildrequestid=41, number=7, masterid=88, builderid=88, workerid=47),
        fakedb.Build(id=31, buildrequestid=42, number=8, masterid=88, builderid=88, workerid=47),
        fakedb.Build(id=40, buildrequestid=43, number=9, masterid=88, builderid=89, workerid=47),
        fakedb.Step(id=131, number=231, name='step231', buildid=30),
        fakedb.Step(id=132, number=232, name='step232', buildid=30),
        fakedb.Step(id=141, number=241, name='step241', buildid=31),
        fakedb.Step(id=142, number=242, name='step242', buildid=40),
    ]

    common_test_result_set_data = [
        fakedb.TestResultSet(
            id=91,
            builderid=88,
            buildid=30,
            stepid=131,
            description='desc1',
            category='cat',
            value_unit='ms',
            tests_failed=None,
            tests_passed=None,
            complete=0,
        ),
        fakedb.TestResultSet(
            id=92,
            builderid=88,
            buildid=30,
            stepid=131,
            description='desc2',
            category='cat',
            value_unit='ms',
            tests_failed=None,
            tests_passed=None,
            complete=1,
        ),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_add_set_get_set(self):
        yield self.db.insert_test_data(self.common_data)
        set_id = yield self.db.test_result_sets.addTestResultSet(
            builderid=88,
            buildid=30,
            stepid=131,
            description='desc',
            category='cat',
            value_unit='ms',
        )
        set_dict = yield self.db.test_result_sets.getTestResultSet(set_id)
        self.assertIsInstance(set_dict, test_result_sets.TestResultSetModel)
        self.assertEqual(
            set_dict,
            test_result_sets.TestResultSetModel(
                id=set_id,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=False,
            ),
        )

    @defer.inlineCallbacks
    def test_get_sets(self):
        yield self.db.insert_test_data([
            *self.common_data,
            fakedb.TestResultSet(
                id=91,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc1',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=0,
            ),
            fakedb.TestResultSet(
                id=92,
                builderid=89,
                buildid=40,
                stepid=142,
                description='desc2',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=1,
            ),
            fakedb.TestResultSet(
                id=93,
                builderid=88,
                buildid=31,
                stepid=141,
                description='desc3',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=1,
            ),
            fakedb.TestResultSet(
                id=94,
                builderid=88,
                buildid=30,
                stepid=132,
                description='desc4',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=1,
            ),
            fakedb.TestResultSet(
                id=95,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc4',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=0,
            ),
        ])

        set_dicts = yield self.db.test_result_sets.getTestResultSets()
        self.assertEqual([d.id for d in set_dicts], [91, 92, 93, 94, 95])
        for d in set_dicts:
            self.assertIsInstance(d, test_result_sets.TestResultSetModel)

        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88)
        self.assertEqual([d.id for d in set_dicts], [91, 93, 94, 95])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=89)
        self.assertEqual([d.id for d in set_dicts], [92])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88, buildid=30)
        self.assertEqual([d.id for d in set_dicts], [91, 94, 95])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88, buildid=31)
        self.assertEqual([d.id for d in set_dicts], [93])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88, stepid=131)
        self.assertEqual([d.id for d in set_dicts], [91, 95])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88, stepid=132)
        self.assertEqual([d.id for d in set_dicts], [94])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88, complete=True)
        self.assertEqual([d.id for d in set_dicts], [93, 94])
        set_dicts = yield self.db.test_result_sets.getTestResultSets(builderid=88, complete=False)
        self.assertEqual([d.id for d in set_dicts], [91, 95])

    @defer.inlineCallbacks
    def test_get_set_from_data(self):
        yield self.db.insert_test_data(self.common_data + self.common_test_result_set_data)

        set_dict = yield self.db.test_result_sets.getTestResultSet(91)
        self.assertEqual(
            set_dict,
            test_result_sets.TestResultSetModel(
                id=91,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc1',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=False,
            ),
        )

    @defer.inlineCallbacks
    def test_get_non_existing_set(self):
        set_dict = yield self.db.test_result_sets.getTestResultSet(91)
        self.assertEqual(set_dict, None)

    @defer.inlineCallbacks
    def test_complete_already_completed_set(self):
        yield self.db.insert_test_data(self.common_data + self.common_test_result_set_data)
        with self.assertRaises(test_result_sets.TestResultSetAlreadyCompleted):
            yield self.db.test_result_sets.completeTestResultSet(92)
        self.flushLoggedErrors(test_result_sets.TestResultSetAlreadyCompleted)

    @async_to_deferred
    async def test_get_test_result_sets_for_commits(self) -> None:
        master_id = fakedb.FakeDBConnector.MASTER_ID
        await self.db.insert_test_data([
            fakedb.Master(id=master_id),
            fakedb.Worker(id=47, name='linux'),
            fakedb.Project(id=100),
            fakedb.Codebase(id=200, projectid=100),
            fakedb.CodebaseCommit(id=300, codebaseid=200, revision='rev300'),
            fakedb.CodebaseCommit(id=301, codebaseid=200, revision='rev301', parent_commitid=300),
            fakedb.CodebaseCommit(id=302, codebaseid=200, revision='rev302', parent_commitid=301),
            fakedb.CodebaseCommit(id=303, codebaseid=200, revision='rev303', parent_commitid=302),
            fakedb.CodebaseCommit(id=304, codebaseid=200, revision='rev304', parent_commitid=303),
            fakedb.Buildset(id=4000),
            fakedb.Buildset(id=4010),
            fakedb.Buildset(id=4020),
            fakedb.Buildset(id=4021),
            fakedb.Buildset(id=4040),
            fakedb.SourceStamp(id=5000, revision='rev300'),
            fakedb.SourceStamp(id=5010, revision='rev301'),
            fakedb.SourceStamp(id=5020, revision='rev302'),
            fakedb.SourceStamp(id=5040, revision='rev304'),
            fakedb.BuildsetSourceStamp(id=6000, buildsetid=4000, sourcestampid=5000),
            fakedb.BuildsetSourceStamp(id=6001, buildsetid=4010, sourcestampid=5010),
            fakedb.BuildsetSourceStamp(id=6002, buildsetid=4020, sourcestampid=5020),
            fakedb.BuildsetSourceStamp(id=6003, buildsetid=4021, sourcestampid=5020),
            fakedb.BuildsetSourceStamp(id=6004, buildsetid=4040, sourcestampid=5040),
            fakedb.Builder(id=400, name='b1'),
            fakedb.BuildRequest(id=7000, buildsetid=4000, builderid=400),
            fakedb.BuildRequest(id=7010, buildsetid=4010, builderid=400),
            fakedb.BuildRequest(id=7020, buildsetid=4020, builderid=400),
            fakedb.BuildRequest(id=7021, buildsetid=4021, builderid=400),
            fakedb.BuildRequest(id=7040, buildsetid=4040, builderid=400),
            fakedb.Build(
                id=8000, buildrequestid=7000, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8010, buildrequestid=7010, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8020, buildrequestid=7020, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8021, buildrequestid=7021, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Build(
                id=8040, buildrequestid=7040, masterid=master_id, builderid=400, workerid=47
            ),
            fakedb.Step(id=9000, buildid=8000),
            fakedb.Step(id=9010, buildid=8010),
            fakedb.Step(id=9020, buildid=8020),
            fakedb.Step(id=9021, buildid=8021),
            fakedb.Step(id=9040, buildid=8040),
            fakedb.TestResultSet(id=10000, builderid=400, buildid=8000, stepid=9000),
            fakedb.TestResultSet(id=10010, builderid=400, buildid=8010, stepid=9010),
            fakedb.TestResultSet(id=10020, builderid=400, buildid=8020, stepid=9020),
            fakedb.TestResultSet(id=10021, builderid=400, buildid=8021, stepid=9021),
            fakedb.TestResultSet(id=10040, builderid=400, buildid=8040, stepid=9040),
        ])

        sets = await self.db.test_result_sets.get_test_result_sets_for_commits(
            commit_ids=[300, 301, 302, 303, 304]
        )
        self.assertEqual(sorted([set.id for set in sets]), [10000, 10010, 10020, 10021, 10040])

    @defer.inlineCallbacks
    def test_complete_set_with_test_counts(self):
        yield self.db.insert_test_data(self.common_data + self.common_test_result_set_data)

        yield self.db.test_result_sets.completeTestResultSet(91, tests_passed=12, tests_failed=2)

        set_dict = yield self.db.test_result_sets.getTestResultSet(91)
        self.assertEqual(
            set_dict,
            test_result_sets.TestResultSetModel(
                id=91,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc1',
                category='cat',
                value_unit='ms',
                tests_failed=2,
                tests_passed=12,
                complete=True,
            ),
        )

    @defer.inlineCallbacks
    def test_complete_set_without_test_counts(self):
        yield self.db.insert_test_data(self.common_data + self.common_test_result_set_data)

        yield self.db.test_result_sets.completeTestResultSet(91)

        set_dict = yield self.db.test_result_sets.getTestResultSet(91)
        self.assertEqual(
            set_dict,
            test_result_sets.TestResultSetModel(
                id=91,
                builderid=88,
                buildid=30,
                stepid=131,
                description='desc1',
                category='cat',
                value_unit='ms',
                tests_failed=None,
                tests_passed=None,
                complete=True,
            ),
        )



# Copyright (C) 2008-2009 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""\
bzr buildbot integration
========================

This file contains both bzr commit/change hooks and a bzr poller.

------------
Requirements
------------

This has been tested with buildbot 0.7.9, bzr 1.10, and Twisted 8.1.0.  It
should work in subsequent releases.

For the hook to work, Twisted must be installed in the same Python that bzr
uses.

-----
Hooks
-----

To install, put this file in a bzr plugins directory (e.g.,
~/.bazaar/plugins). Then, in one of your bazaar conf files (e.g.,
~/.bazaar/locations.conf), set the location you want to connect with buildbot
with these keys:

- buildbot_on: one of 'commit', 'push, or 'change'.  Turns the plugin on to
  report changes via commit, changes via push, or any changes to the trunk.
  'change' is recommended.

- buildbot_server: (required to send to a buildbot master) the URL of the
  buildbot master to which you will connect (as of this writing, the same
  server and port to which workers connect).

- buildbot_port: (optional, defaults to 9989) the port of the buildbot master
  to which you will connect (as of this writing, the same server and port to
  which workers connect)

- buildbot_auth: (optional, defaults to change:changepw) the credentials
  expected by the change source configuration in the master. Takes the
  "user:password" form.

- buildbot_pqm: (optional, defaults to not pqm) Normally, the user that
  commits the revision is the user that is responsible for the change.  When
  run in a pqm (Patch Queue Manager, see https://launchpad.net/pqm)
  environment, the user that commits is the Patch Queue Manager, and the user
  that committed the *parent* revision is responsible for the change.  To turn
  on the pqm mode, set this value to any of (case-insensitive) "Yes", "Y",
  "True", or "T".

- buildbot_dry_run: (optional, defaults to not a dry run) Normally, the
  post-commit hook will attempt to communicate with the configured buildbot
  server and port. If this parameter is included and any of (case-insensitive)
  "Yes", "Y", "True", or "T", then the hook will simply print what it would
  have sent, but not attempt to contact the buildbot master.

- buildbot_send_branch_name: (optional, defaults to not sending the branch
  name) If your buildbot's bzr source build step uses a repourl, do
  *not* turn this on. If your buildbot's bzr build step uses a baseURL, then
  you may set this value to any of (case-insensitive) "Yes", "Y", "True", or
  "T" to have the buildbot master append the branch name to the baseURL.

Note: The bzr smart server (as of version 2.2.2) doesn't know how to resolve
bzr:// urls into absolute paths so any paths in locations.conf won't match,
hence no change notifications will be sent to Buildbot. Setting configuration
parameters globally or in-branch might still work.

When buildbot no longer has a hardcoded password, it will be a configuration
option here as well.

------
Poller
------

See the Buildbot manual.

-------------------
Contact Information
-------------------

Maintainer/author: gary.poster@canonical.com
"""

# Work around Twisted bug.
# See http://twistedmatrix.com/trac/ticket/3591
import socket
from typing import ClassVar
from typing import Sequence

import bzrlib.branch
import bzrlib.errors
import bzrlib.trace
import twisted.cred.credentials
import twisted.internet.base
import twisted.internet.reactor
import twisted.internet.selectreactor
import twisted.internet.task
import twisted.internet.threads
import twisted.python.log
import twisted.spread.pb
from twisted.internet import defer
from twisted.python import failure

try:
    import buildbot.changes.base
    import buildbot.changes.changes
    import buildbot.util
except ImportError:
    DEFINE_POLLER = False
else:
    DEFINE_POLLER = True


#
# This is the code that the poller and the hooks share.


def generate_change(
    branch, old_revno=None, old_revid=None, new_revno=None, new_revid=None, blame_merge_author=False
):
    """Return a dict of information about a change to the branch.

    Dict has keys of "files", "who", "comments", and "revision", as used by
    the buildbot Change (and the PBChangeSource).

    If only the branch is given, the most recent change is returned.

    If only the new_revno is given, the comparison is expected to be between
    it and the previous revno (new_revno -1) in the branch.

    Passing old_revid and new_revid is only an optimization, included because
    bzr hooks usually provide this information.

    blame_merge_author means that the author of the merged branch is
    identified as the "who", not the person who committed the branch itself.
    This is typically used for PQM.
    """
    change = {}  # files, who, comments, revision; NOT branch (= branch.nick)
    if new_revno is None:
        new_revno = branch.revno()
    if new_revid is None:
        new_revid = branch.get_rev_id(new_revno)
    # TODO: This falls over if this is the very first revision
    if old_revno is None:
        old_revno = new_revno - 1
    if old_revid is None:
        old_revid = branch.get_rev_id(old_revno)
    repository = branch.repository
    new_rev = repository.get_revision(new_revid)
    if blame_merge_author:
        # this is a pqm commit or something like it
        change['who'] = repository.get_revision(new_rev.parent_ids[-1]).get_apparent_authors()[0]
    else:
        change['who'] = new_rev.get_apparent_authors()[0]
    # maybe useful to know:
    # name, email = bzrtools.config.parse_username(change['who'])
    change['comments'] = new_rev.message
    change['revision'] = new_revno
    files = change['files'] = []
    changes = repository.revision_tree(new_revid).changes_from(repository.revision_tree(old_revid))
    for collection, name in (
        (changes.added, 'ADDED'),
        (changes.removed, 'REMOVED'),
        (changes.modified, 'MODIFIED'),
    ):
        for info in collection:
            path = info[0]
            kind = info[2]
            files.append(' '.join([path, kind, name]))
    for info in changes.renamed:
        oldpath, newpath, id, kind, text_modified, meta_modified = info
        elements = [oldpath, kind, 'RENAMED', newpath]
        if text_modified or meta_modified:
            elements.append('MODIFIED')
        files.append(' '.join(elements))
    return change


#
# poller

# We don't want to make the hooks unnecessarily depend on buildbot being
# installed locally, so we conditionally create the BzrPoller class.
if DEFINE_POLLER:
    FULL = object()
    SHORT = object()

    class BzrPoller(buildbot.changes.base.PollingChangeSource, buildbot.util.ComparableMixin):
        compare_attrs: ClassVar[Sequence[str]] = 'url'

        def __init__(
            self,
            url,
            poll_interval=10 * 60,
            blame_merge_author=False,
            branch_name=None,
            category=None,
        ):
            # poll_interval is in seconds, so default poll_interval is 10
            # minutes.
            # bzr+ssh://bazaar.launchpad.net/~launchpad-pqm/launchpad/devel/
            # works, lp:~launchpad-pqm/launchpad/devel/ doesn't without help.
            if url.startswith('lp:'):
                url = 'bzr+ssh://bazaar.launchpad.net/' + url[3:]
            self.url = url
            self.poll_interval = poll_interval
            self.loop = twisted.internet.task.LoopingCall(self.poll)
            self.blame_merge_author = blame_merge_author
            self.branch_name = branch_name
            self.category = category

        def startService(self):
            twisted.python.log.msg(f"BzrPoller({self.url}) starting")
            if self.branch_name is FULL:
                ourbranch = self.url
            elif self.branch_name is SHORT:
                # We are in a bit of trouble, as we cannot really know what our
                # branch is until we have polled new changes.
                # Seems we would have to wait until we polled the first time,
                # and only then do the filtering, grabbing the branch name from
                # whatever we polled.
                # For now, leave it as it was previously (compare against
                # self.url); at least now things work when specifying the
                # branch name explicitly.
                ourbranch = self.url
            else:
                ourbranch = self.branch_name
            for change in reversed(self.parent.changes):
                if change.branch == ourbranch:
                    self.last_revision = change.revision
                    break
            else:
                self.last_revision = None
            buildbot.changes.base.PollingChangeSource.startService(self)

        def stopService(self):
            twisted.python.log.msg(f"BzrPoller({self.url}) shutting down")
            return buildbot.changes.base.PollingChangeSource.stopService(self)

        def describe(self):
            return f"BzrPoller watching {self.url}"

        @defer.inlineCallbacks
        def poll(self):
            # On a big tree, even individual elements of the bzr commands
            # can take awhile. So we just push the bzr work off to a
            # thread.
            try:
                changes = yield twisted.internet.threads.deferToThread(self.getRawChanges)
            except (SystemExit, KeyboardInterrupt):
                raise
            except Exception:
                # we'll try again next poll.  Meanwhile, let's report.
                twisted.python.log.err()
            else:
                for change_kwargs in changes:
                    yield self.addChange(change_kwargs)
                    self.last_revision = change_kwargs['revision']

        def getRawChanges(self):
            branch = bzrlib.branch.Branch.open_containing(self.url)[0]
            if self.branch_name is FULL:
                branch_name = self.url
            elif self.branch_name is SHORT:
                branch_name = branch.nick
            else:  # presumably a string or maybe None
                branch_name = self.branch_name
            changes = []
            change = generate_change(branch, blame_merge_author=self.blame_merge_author)
            if self.last_revision is None or change['revision'] > self.last_revision:
                change['branch'] = branch_name
                change['category'] = self.category
                changes.append(change)
                if self.last_revision is not None:
                    while self.last_revision + 1 < change['revision']:
                        change = generate_change(
                            branch,
                            new_revno=change['revision'] - 1,
                            blame_merge_author=self.blame_merge_author,
                        )
                        change['branch'] = branch_name
                        changes.append(change)
            changes.reverse()
            return changes

        def addChange(self, change_kwargs):
            d = defer.Deferred()

            def _add_change():
                d.callback(self.master.data.updates.addChange(src='bzr', **change_kwargs))

            twisted.internet.reactor.callLater(0, _add_change)
            return d

#
# hooks

HOOK_KEY = 'buildbot_on'
SERVER_KEY = 'buildbot_server'
PORT_KEY = 'buildbot_port'
AUTH_KEY = 'buildbot_auth'
DRYRUN_KEY = 'buildbot_dry_run'
PQM_KEY = 'buildbot_pqm'
SEND_BRANCHNAME_KEY = 'buildbot_send_branch_name'

PUSH_VALUE = 'push'
COMMIT_VALUE = 'commit'
CHANGE_VALUE = 'change'


def _is_true(config, key):
    val = config.get_user_option(key)
    return val is not None and val.lower().strip() in ('y', 'yes', 't', 'true')


def _installed_hook(branch):
    value = branch.get_config().get_user_option(HOOK_KEY)
    if value is not None:
        value = value.strip().lower()
        if value not in (PUSH_VALUE, COMMIT_VALUE, CHANGE_VALUE):
            raise bzrlib.errors.BzrError(
                f'{HOOK_KEY}, if set, must be one of {PUSH_VALUE}, {COMMIT_VALUE}, or {CHANGE_VALUE}'
            )
    return value


# replaces twisted.internet.thread equivalent


def _putResultInDeferred(reactor, deferred, f, args, kwargs):
    """
    Run a function and give results to a Deferred.
    """
    try:
        result = f(*args, **kwargs)
    except Exception:
        f = failure.Failure()
        reactor.callFromThread(deferred.errback, f)
    else:
        reactor.callFromThread(deferred.callback, result)


# would be a proposed addition.  deferToThread could use it


def deferToThreadInReactor(reactor, f, *args, **kwargs):
    """
    Run function in thread and return result as Deferred.
    """
    d = defer.Deferred()
    reactor.callInThread(_putResultInDeferred, reactor, d, f, args, kwargs)
    return d


# uses its own reactor for the threaded calls, unlike Twisted's


class ThreadedResolver(twisted.internet.base.ThreadedResolver):
    def getHostByName(self, name, timeout=(1, 3, 11, 45)):
        if timeout:
            timeoutDelay = sum(timeout)
        else:
            timeoutDelay = 60
        userDeferred = defer.Deferred()
        lookupDeferred = deferToThreadInReactor(self.reactor, socket.gethostbyname, name)
        cancelCall = self.reactor.callLater(timeoutDelay, self._cleanup, name, lookupDeferred)
        self._runningQueries[lookupDeferred] = (userDeferred, cancelCall)
        lookupDeferred.addBoth(self._checkTimeout, name, lookupDeferred)
        return userDeferred


def send_change(branch, old_revno, old_revid, new_revno, new_revid, hook):
    config = branch.get_config()
    server = config.get_user_option(SERVER_KEY)
    if not server:
        bzrlib.trace.warning(
            'bzr_buildbot: ERROR.  If %s is set, %s must be set', HOOK_KEY, SERVER_KEY
        )
        return
    change = generate_change(
        branch,
        old_revno,
        old_revid,
        new_revno,
        new_revid,
        blame_merge_author=_is_true(config, PQM_KEY),
    )
    if _is_true(config, SEND_BRANCHNAME_KEY):
        change['branch'] = branch.nick
    # as of this writing (in Buildbot 0.7.9), 9989 is the default port when
    # you make a buildbot master.
    port = int(config.get_user_option(PORT_KEY) or 9989)
    # if dry run, stop.
    if _is_true(config, DRYRUN_KEY):
        bzrlib.trace.note(
            "bzr_buildbot DRY RUN (*not* sending changes to %s:%d on %s)", server, port, hook
        )
        keys = sorted(change.keys())
        for k in keys:
            bzrlib.trace.note("[%10s]: %s", k, change[k])
        return
    # We instantiate our own reactor so that this can run within a server.
    reactor = twisted.internet.selectreactor.SelectReactor()
    # See other reference to http://twistedmatrix.com/trac/ticket/3591
    # above.  This line can go away with a release of Twisted that addresses
    # this issue.
    reactor.resolver = ThreadedResolver(reactor)
    pbcf = twisted.spread.pb.PBClientFactory()
    reactor.connectTCP(server, port, pbcf)
    auth = config.get_user_option(AUTH_KEY)
    if auth:
        user, passwd = [s.strip() for s in auth.split(':', 1)]
    else:
        user, passwd = ('change', 'changepw')
    deferred = pbcf.login(twisted.cred.credentials.UsernamePassword(user, passwd))

    @deferred.addCallback
    def sendChanges(remote):
        """Send changes to buildbot."""
        bzrlib.trace.mutter("bzrbuildout sending changes: %s", change)
        change['src'] = 'bzr'
        return remote.callRemote('addChange', change)

    def quit(ignore, msg):
        bzrlib.trace.note("bzrbuildout: %s", msg)
        reactor.stop()

    deferred.addCallback(quit, "SUCCESS")

    @deferred.addErrback
    def failed(failure):
        bzrlib.trace.warning("bzrbuildout: FAILURE\n %s", failure)
        reactor.stop()

    reactor.callLater(60, quit, None, "TIMEOUT")
    bzrlib.trace.note(
        "bzr_buildbot: SENDING CHANGES to buildbot master %s:%d on %s", server, port, hook
    )
    reactor.run(installSignalHandlers=False)  # run in a thread when in server


def post_commit(
    local_branch,
    master_branch,  # branch is the master_branch
    old_revno,
    old_revid,
    new_revno,
    new_revid,
):
    if _installed_hook(master_branch) == COMMIT_VALUE:
        send_change(master_branch, old_revid, old_revid, new_revno, new_revid, COMMIT_VALUE)


def post_push(result):
    if _installed_hook(result.target_branch) == PUSH_VALUE:
        send_change(
            result.target_branch,
            result.old_revid,
            result.old_revid,
            result.new_revno,
            result.new_revid,
            PUSH_VALUE,
        )


def post_change_branch_tip(result):
    if _installed_hook(result.branch) == CHANGE_VALUE:
        send_change(
            result.branch,
            result.old_revid,
            result.old_revid,
            result.new_revno,
            result.new_revid,
            CHANGE_VALUE,
        )


bzrlib.branch.Branch.hooks.install_named_hook(
    'post_commit', post_commit, 'send change to buildbot master'
)
bzrlib.branch.Branch.hooks.install_named_hook(
    'post_push', post_push, 'send change to buildbot master'
)
bzrlib.branch.Branch.hooks.install_named_hook(
    'post_change_branch_tip', post_change_branch_tip, 'send change to buildbot master'
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from unittest import mock

from twisted.trial import unittest

from buildbot.test.fake import fakeprotocol
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import protocols
from buildbot.worker.protocols import base


class TestFakeConnection(protocols.ConnectionInterfaceTest, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        self.worker = mock.Mock()
        self.conn = fakeprotocol.FakeConnection(self.worker)


class TestConnection(protocols.ConnectionInterfaceTest, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        self.worker = mock.Mock()
        self.conn = base.Connection(self.worker.workername)

    def test_notify(self):
        cb = mock.Mock()

        self.conn.notifyOnDisconnect(cb)
        self.assertEqual(cb.call_args_list, [])
        self.conn.notifyDisconnected()
        self.assertNotEqual(cb.call_args_list, [])
